/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <stdint.h>
#ifndef WIN32
#include <libgen.h>
#endif
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <cctype>
#include <cstdio>
#include <iostream>
#include <cstring>
#include <string>
#include <cstdlib>
#include <new>

#include "../logic/defines.h"
#include "../logic/util.h"
#include "../logic/dict.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/game.h"

#ifdef WIN32
# define random rand
# define srandom srand
#endif

#ifdef WIN32
# define ANSI_U ""
# define ANSI_B ""
# define ANSI_BU ""
# define ANSI_E ""
#else
# define ANSI_U "\033[4m"
# define ANSI_B "\033[1m"
# define ANSI_BU "\033[1;4m"
# define ANSI_E "\033[m"
#endif

using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::string;

static void	solveit(const char *letters);
static void	playit(int min, int max);

static void	print_game(const char *word, int wpl);
static void	play_loop(const char *w);
static int	wpl_calc(int w);

static void	help(void);
static void	instr(void);

char	dictpath[PATH_MAX];
bool	plural_sieve = PLURAL_SIEVE;
bool	precomp = PRECOMP;
const int FW = 9;

Spellathon::Dict	*dict;
Spellathon::Game	*game;
Spellathon::DispWordsBuf *dsp;

int
main(int argc, char *argv[])
{
	char	letters[MAX_LETRS+1];
	char	*tok = 0;
	int		i, ch, dflag=0, sflag=0, pflag=0, cflag=0, eflag=0;
	int 	minsol = MIN_SOL, maxsol = MAX_SOL;

#ifndef WIN32
	if (chdir(dirname(argv[0])) < 0)
		Spellathon::spth_error(true, "Can't change directory to: %s", argv[0]);
#endif

	if (argc < 2)
		help();
	while ((ch = getopt(argc, argv, "hc:ie:d:pr:s:")) != -1) {
		switch (ch) {
		case 'h':
			help();
			break;

		case 'i':
			instr();
			break;

		case 'r':
			minsol = maxsol = 0;
			minsol = strtol(optarg, 0, 10);
			if (minsol > 0) tok = strchr(optarg, ',');
			if (tok)	maxsol = strtol(tok+1, 0, 10);
			if (!maxsol || maxsol - minsol < MGAP_SOL) {
				cerr << "Invalid solution range."
					" (Compiled limit: " 
				     << MMIN_SOL << "," << MMAX_SOL << ")\n";
				exit(1);
			}
			break;

		case 's':
			sflag++;
			strncpy(letters, optarg, MAX_LETRS);
			letters[MAX_LETRS] = 0;
			for (i = 0; i < MAX_LETRS; i++) {
				if (!isalpha(letters[i])) {
					cerr << "Please enter seven letters.\n";
					exit(1);
				}
				letters[i] = tolower(letters[i]);
			}
			Spellathon::swapch(&letters[0], &letters[MAX_LETRS-1]);
			break;

		case 'p':
			pflag++;
			break;

		case 'c':
			cflag++;
			if (strcmp(optarg, "on") == 0) {
				precomp = true;
			} else if (strcmp(optarg, "off") == 0) {
				precomp = false;
			} else {
				cerr << "invalid argument for '-c'" << endl;
			}
			break;

		case 'd':
			dflag++;
			strncpy(dictpath, optarg, sizeof(dictpath)-1);
			dictpath[sizeof(dictpath)-1] = 0;
			break;

		case 'e':
			eflag++;
			if (strcmp(optarg, "on") == 0) {
				plural_sieve = true;
			} else if (strcmp(optarg, "off") == 0) {
				plural_sieve = false;
			} else {
				cerr << "invalid argument for '-e'" << endl;
			}
			break;

		case '?':
		default:
			help();
			break;
		}
	} 
	argc -= optind;
	argv += optind;

	if (sflag && pflag) 
		help();

	if (!dflag) 
		strcpy(dictpath, DICTPATH);

	try {
		dict = new Spellathon::Dict;
		dict->init(dictpath, precomp, plural_sieve);
		game = new Spellathon::Game(dict, plural_sieve);
		dsp = new Spellathon::DispWordsBuf;
	}
	catch (string s)
	{
		cerr << s << endl;
		exit(1);
	}
	catch (std::bad_alloc &ba)
	{
		cerr << "Can't allocate memory: " << ba.what() << endl;
		exit(1);
	}

	if (sflag) 
		solveit(letters);

	if (pflag) {
		srandom(time(0));
		playit(minsol, maxsol);
	}

	delete dsp;
	delete game;
	delete dict;

	return 0;
}

static void
playit(int minsol, int maxsol)
{
	bool r = game->play(precomp, minsol, maxsol);
	if (r) {
		dsp->init(game->words(), wpl_calc(game->nWords()));
		play_loop(game->puzzle().c_str());
	} else {
		cout <<  "No suitable game found. "
		       	 "Choose a different Solution range or use "
				 "a different dictionary.\n";
	}
}
	
static void
play_loop(const char *w)
{
	int wpl = dsp->wpl();

	time_t t = time(0);
	string op = "";
	int fw = wpl*FW - sizeof("words:   /  ");

	while (1) {
		print_game(w, wpl);

		if (op != "")
			op = "[ " + op + " ]";
		printf("%s" ANSI_B "\n%s%*swords: %d/%d\n\n" ANSI_E
				"Enter a word ('f' to finish, 'h' for hint):\n"
				, dsp->ansiColor(),  op.c_str(), fw-(int)op.size()," ", 
				dsp->nSetWords(), dsp->nWords()
		);

		string word;
		cin >> word;

		if (word == "h") {
			if (!dsp->setHint()) 
				op = "No more hints";
			else
				op = "";

			continue;
		}

		if (word == "f") {
fin:
			time_t t2 = time(0);

			print_game(w, wpl);
			dsp->setSolved();
			printf("%s\n\n", dsp->ansiColor());
			printf(ANSI_B "Correct words: %d/%d    " ANSI_E,
					dsp->nSetWords(), dsp->nWords()); 

			game->calcScore(dsp->nSetWords(), dsp->nWords(), 
					 dsp->hintLevel(), t2-t, dsp->entered7());

			string rating = game->rating() == "" ? "" : "[ " + game->rating() + " ]";
			printf(ANSI_B "Score: %d %s    Time: %02d:%02d\n\n" ANSI_E,
				(int)game->score(), rating.c_str(),
				(int)(t2-t)/60, (int)(t2-t)%60);

			return;
		}

		if (word.size() > 7 || word.size() < 4) {
				op = "No match";
		}

		switch (dsp->setWord(word)) {
		case 1:
			op = "";
			if (dsp->solved()) 
				goto fin;			
			break;
		case -1:
			op = "No match";
			break;
		case 0:
			op = "Already entered";
			break;
		}
	}
}

static void
print_game(const char *word, int wpl)
{
	int lm = wpl * FW / 2;
	if (lm >= 3)
		lm -= 3;

	printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n%*c\n\n"
		"%*c     %c\n\n"
		"%*c" ANSI_BU "%c\n\n" ANSI_E
		"%*c     %c\n\n"
		"%*c\n\n"
		, lm+3, toupper(word[1])
		, lm, toupper(word[2]), toupper(word[3])
		, lm+2, ' ', toupper(word[6])
		, lm, toupper(word[4]), toupper(word[5])
		, lm+3, toupper(word[0])
	);
}

static int
wpl_calc(int words)
{
	int wpl;
	const int max = 8;

	if (words > 32) {
		wpl = (words + (max-1)) / max;
		if (wpl > max)
			wpl = max;
	} else {
		wpl = 4;
	}

	return wpl;
}

static void
solveit(const char *letters)
{
	int r = game->solve(letters);
	if (r) {
		dsp->init(game->words(), wpl_calc(game->nWords()));
		dsp->setSolved();
		cout << endl << dsp->ansiColor() << endl;
	} 
	cout << dsp->nWords() << " words\n\n";
}

static void	
help(void)
{
	const char* svn_version(void);
	fprintf(stderr, "\nSpellathon "VERSION" (svn %s)    Copyright (c) 2009 Girish Rao\n\n"
		"spellathon < -p | -s <string> | -i > [ -r min,max ] [ -d /path/to/dict ] [ -c on|off ] [ -e on|off ]\n\n"
		"Main options:\n\n"
		"  -i  Game instructions.\n"
		"  -p  Play a new game.\n"
        "  -s  Solve a custom puzzle.\n"
		"      Enter the central letter followed by\n"
        "       the remaining letters.\n"
		"      Example: spellathon -s taplerss (central letter is 't').\n\n"
		"Sub options:\n\n"
		"  -r  Play games with #solutions in the range [min,max].  (default: %d,%d)\n"
		"  -c  Use precomputed dictionary.  (default: %s)\n"
		"  -d  Use specified dictionary.  (default: '%s')\n"
		"  -e  Sieve plurals.  (default: on)\n\n",
		svn_version(), MIN_SOL, MAX_SOL, PRECOMP ? "on":"off", DICTPATH);
	exit(1);
}

static void	
instr(void)
{
	printf("\n" ANSI_U "Game instructions" ANSI_E "\n\n"
		"  o Find all the words of four or more letters from the\n"
		"     letters shown in the puzzle.\n  o In making a word, each"
		" letter may be used once only.\n  o Each word must contain"
		" the central letter.\n  o There should be atleast one seven"
        " letter word.\n  o Plurals, foreign words and proper names are not allowed.\n\n"
		ANSI_U "Notes" ANSI_E"\n\n  - Using hints will reduce your score.\n"
		"  - To get the top score/rating you will need to be fast.\n"
		"  - The seven letter word(s) is worth extra points.\n"
		"  - Ratings: %s, %s, %s.\n\n", Ratings[1], Ratings[2], Ratings[3]
	);
	exit(0);
}
