/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
# include <sys/stat.h>
# include <pwd.h>
#endif /* WIN32 */

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>

#ifndef SPTH_CL
# include <QProgressDialog>
# include <QFileInfo>
# include <QDateTime>
#endif /* SPTH_CL */

#include "dict.h"
#include "util.h"
#include "game.h"

namespace Spellathon
{

	static int dictSortcmp(const void *p1, const void *p2);
	static int precompSortcmp(const void *p1, const void *p2);
	static int wordSortcmp(const void *p1, const void *p2);

	void
	Dict::dictOpen(const char *path)
	{
		char	line[MAX_LNLEN], *r; 
		int		nelem, slen, i;
		FILE	*fd;
	
		fd = fopen(path, "r");
		if (fd == 0) 
			spth_error(1, "Can't open dictionary: '%s'", path);
	
		nent_ = 0;
		nelem = 1<<14;
	
		dictarr = (dict_ent *)memalloc(nelem * sizeof(dict_ent));
		memset(dictarr, 0, nelem*sizeof(dict_ent));
	
		/* 
		 * read the wordlist and find all (appropriate) 
		 * words and store their signatures in mem.
		 */
again:
		while ( (r = fgets(line, MAX_LNLEN, fd)) != NULL) {
			slen = strlen(line);

			if (line[slen-1] != '\n') {
				fclose(fd);
				destroy();
				errno = EINVAL;
				spth_error(0, "'%s': Unrecognized file type. Dictionary must "
						   "be an ascii text file with the words seperated by newlines."
						   , path);
			}

			/* Remove newline char */
			line[--slen] = 0;

			/* Remove /r if file is in Windows format */
			if (line[slen-1] == '\r')
				line[--slen] = 0;
	
			if (slen < MIN_LETRS-1 || slen > MAX_LETRS)
				goto again;
	
			for (i = 0; i < slen; i++) 
				if(!islower(line[i]))
					goto again;
	
			if (nent_ >= nelem) {
				dictarr = (dict_ent *)memresize(dictarr, nelem * sizeof(dict_ent) * 2);
				memset(dictarr + nelem, 0, nelem * sizeof(dict_ent));
				nelem *= 2;
			}
	
			strcpy(dictarr[nent_].word,line);
			sign(line, slen);
			strcpy(dictarr[nent_++].sign,line);
		}
	
		if (ferror(fd)) {
			fclose(fd);
			destroy();
			spth_error(1, "'%s': Read error", path);
		}
	
		fclose(fd);
	
		/* sort the dictionary */
		qsort((void *)dictarr, nent_, sizeof(dict_ent), dictSortcmp);
	
		words = (char (*)[8])memalloc(nent_ * 8);
		memset(words, 0, nent_ * 8);
		for (int i = 0; i < nent_; i++) 
			strcpy(words[i], dictarr[i].word);

		qsort((void *)words, nent_, 8, wordSortcmp);

		/* find strlen offsets */
	
		memset(off, 0, sizeof(off)); 

		off[3].start = off[3].end = -1;
		off[4].start = off[4].end = -1;
		off[5].start = off[5].end = -1;
		off[6].start = off[6].end = -1;
		off[7].start = off[7].end = -1;
		int curlen = 0;

		for (int i = 0; i < nent_; i++) {
			int len = strlen(dictarr[i].sign);
			if (len > curlen) {
				off[len].start = i;
				curlen = len;
			}
			off[curlen].end = i;
		}
	}

#ifdef SPTH_CL
	Dict::Dict() 
		: dictarr(0), precomparr(0), words(0)
#else
	Dict::Dict(QWidget* p) 
		: dictarr(0), precomparr(0), words(0), gui_h_(p)
#endif /* SPTH_CL */
	{ 
		dictbin_[0] = '\0';
	}

	void
	Dict::init(const char *path, bool precomp, bool plural_sieve) 
	{
		dictPath_ = path;
		destroy();
		precomp_ = precomp;
		plural_sieve_ = plural_sieve;
		if (precomp_) {
			setPrecompPath();
			precompOpen(path); 
		} else {
			dictOpen(path); 
		}
	}
	
	void
	Dict::setPrecompPath()
	{
#ifdef WIN32
		strcpy(dictbin_, getenv("APPDATA"));
		strcat(dictbin_, "/"DICTBIN);
#else
		struct passwd *pw = getpwuid(getuid());
		if (pw == 0)
			spth_error(1, "Can't get user info.");

		strcpy(dictbin_, pw->pw_dir);
		strcat(dictbin_, "/.spellathon/");
		if (access(dictbin_, R_OK | W_OK) < 0) {
			if (mkdir(dictbin_, 
			          S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) < 0)
 			       spth_error(1, "Can't create '%s'", dictbin_);
		}
			
		strcat(dictbin_, DICTBIN);
#endif /* WIN32 */
	}

	void
	Dict::precompOpen(const char *path)
	{
		FILE *binfl = fopen(dictbin_, "rb");	
		if (binfl == 0) {
			if (errno == ENOENT) {
				precompCreate(path);
				return;
			} else {
 			       spth_error(1, "Can't open precomputed dictionary: '%s'", dictbin_);
			}
		}
			

		if (fread(&precomp_hdr, sizeof(precomp_hdr_st), 1, binfl) < 
			(unsigned int)1) {
			goto oops;
		}
#ifdef SPTH_CL 
# ifndef WIN32
		struct stat binp, dictp;
		stat(dictbin_, &binp);
		stat(path, &dictp);
# endif
#endif /* SPTH_CL */
		if (strcmp(precomp_hdr.path, path) ||
		    strcmp(precomp_hdr.magic, MAGIC) ||
			(precomp_hdr.plural_sieve > 0) != plural_sieve_
#ifdef SPTH_CL 
# ifndef WIN32
		    || binp.st_mtime < dictp.st_mtime
# endif
#else
		    || QFileInfo(dictbin_).lastModified().toTime_t() <
		    QFileInfo(path).lastModified().toTime_t()
#endif /* SPTH_CL */
		) {
			fclose(binfl);
			precompCreate(path);
			return;
		}

		nent_ = precomp_hdr.dict_nent;
		memcpy(off, precomp_hdr.off, 8 * sizeof(off_st));

		precomparr = (precomp_ent*)memalloc(precomp_hdr.precomp_nent *
			     sizeof(precomp_ent));

		dictarr = (dict_ent*)memalloc(precomp_hdr.dict_nent * sizeof(dict_ent));

		if (fread(dictarr, sizeof(dict_ent), precomp_hdr.dict_nent, binfl) 
		    < (unsigned int)precomp_hdr.dict_nent) {
			goto oops;
		}

		if (fread(precomparr, sizeof(precomp_ent), precomp_hdr.precomp_nent, binfl) 
			< (unsigned int)precomp_hdr.precomp_nent) {
			goto oops;
		}

		words = (char (*)[8])memalloc(nent_ * 8);
		memset(words, 0, nent_ * 8);
		for (int i = 0; i < nent_; i++) 
			strcpy(words[i], dictarr[i].word);

		qsort((void *)words, nent_, 8, wordSortcmp);

		fclose(binfl);
		return;

oops:
		fclose(binfl);
		destroy();
		spth_error(1, "'%s': Read error", dictbin_);
	}

	void
	Dict::precompCreate(const char *path)
	{
		int cnt = 0, n;

		dictOpen(path);
		Game	game(this, plural_sieve_);

		precomparr = (precomp_ent*)memalloc(nent7() * 7 * sizeof(precomp_ent));

#ifndef SPTH_CL
		QProgressDialog pdiag("Precomputing Dictionary ...", 0, 
							  off[7].start , off[7].end, gui_h_);
		pdiag.setWindowModality(Qt::WindowModal);
		pdiag.setMinimumDuration(0);
		pdiag.setWindowFlags(Qt::Popup);
		pdiag.move((gui_h_->size().width()-pdiag.size().width())/2 + gui_h_->pos().x(),
				   gui_h_->size().height()/2 + gui_h_->pos().y());
#endif /* SPTH_CL */

		for (int i = off[7].start; i <= off[7].end; i++) {
			char sig[MAX_LETRS+1];
			strcpy(sig, dictarr[i].sign);

			int k = MAX_LETRS - 2;
			do {
				if ((n = game.solver(sig)) > 0) {
					strcpy(precomparr[cnt].sign, sig);	
					precomparr[cnt++].nsol = n;
				}

				if (k < 0)
					break;

				/* try a different central letter */	
				swapch(&sig[MAX_LETRS-1], &sig[k--]);
			} while (1);

			if (!(i & ((1<<8) - 1))) {
#ifndef SPTH_CL
				pdiag.setValue(i);
#else
				printf("\rPrecomputing dictionary... %d%%", (i-off[7].start)*100/nent7());
				fflush(stdout);
#endif /* SPTH_CL */
			}
		}

#ifdef SPTH_CL
                printf("\rPrecomputing dictionary... done.\n");
#endif /* SPTH_CL */

		/* sort by number of solutions */
		qsort((void *)precomparr, cnt, sizeof(precomp_ent), precompSortcmp);

		/* remove dup signs */
		if (cnt) {
			precomp_ent *f = precomparr;
			precomp_ent *l = precomparr + cnt;
			precomp_ent *r = f;
			while (++f != l) {
				if (strcmp(r->sign, f->sign))
					memmove(++r, f, sizeof(precomp_ent));
			}
			++r;
			cnt = r - precomparr;
		}

		/* fill header */
		memset(&precomp_hdr, 0, sizeof(precomp_hdr_st));
		precomp_hdr.dict_nent = nent_;
		precomp_hdr.precomp_nent = cnt;
		memcpy(precomp_hdr.off, off, 8 * sizeof(off_st));
		strncpy(precomp_hdr.path, path, MAX_PATH_LEN);
		precomp_hdr.path[MAX_PATH_LEN-1] = 0;
		strcpy(precomp_hdr.magic, MAGIC);
		precomp_hdr.plural_sieve = plural_sieve_;
		
		FILE *binfl = fopen(dictbin_, "wb");
		if (binfl == 0) {
			destroy();
			spth_error(1, "Can't open precomputed dictionary: '%s'", dictbin_);
		}

		if (fwrite(&precomp_hdr, sizeof(precomp_hdr_st), 1, binfl) < 1 ||
			fwrite(dictarr, sizeof(dict_ent), precomp_hdr.dict_nent, binfl) < (unsigned int)precomp_hdr.dict_nent ||
			fwrite(precomparr, sizeof(precomp_ent), precomp_hdr.precomp_nent, binfl) < (unsigned int)precomp_hdr.precomp_nent
		   ) {
			fclose(binfl);
			destroy();
			spth_error(1, "'%s': Write error", dictbin_);
		}

		fclose(binfl);
	}

	/* binary search the dict. */
	/* performance critical routine */
	inline int
	Dict::bsearch(const char * sign, int slen) const
	{
		int i, l, h, m, r;
	
		l = off[slen].start;
		h = off[slen].end;
		
		if (l != -1 && h != -1) {
			char *dictsign;
			while (h >= l) {
				m = (l + h) / 2;

				/* r = strcmp(sign, dictarr[m].sign) */
				r = 0;
				i = 0;
				dictsign = dictarr[m].sign; 
				if (sign[0] != dictsign[0]) { goto out; }
				if (sign[1] != dictsign[1]) { i = 1; goto out; }
				if (sign[2] != dictsign[2]) { i = 2; goto out; }
				if (sign[3] != dictsign[3]) { i = 3; goto out; }
				if (sign[4] != dictsign[4]) { i = 4; goto out; }
				if (sign[5] != dictsign[5]) { i = 5; goto out; }
				if (sign[6] != dictsign[6]) { i = 6; }
out:
				r = sign[i] - dictsign[i];

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return m;
			}	
		} 

		return -1;
	}

	/* performance critical routine */
	bool
	Dict::matchWord(const char *word) const
	{
		int i, l, h, m, r;
	
		int slen = strlen(word);
		l = off[slen].start;
		h = off[slen].end;
		
		if (l != -1 && h != -1) {
			while (h >= l) {
				m = (l + h) / 2;

				r = 0;
				i = 0;
				if (word[0] != words[m][0]) { goto out; }
				if (word[1] != words[m][1]) { i = 1; goto out; }
				if (word[2] != words[m][2]) { i = 2; goto out; }
				if (word[3] != words[m][3]) { i = 3; goto out; }
				if (word[4] != words[m][4]) { i = 4; goto out; }
				if (word[5] != words[m][5]) { i = 5; goto out; }
				if (word[6] != words[m][6]) { i = 6; }
out:
				r = word[i] - words[m][i];

				if (r < 0)
						h = m - 1;
				else if (r > 0)
						l = m + 1;
				else
					return true;
			}	
		} 

		return false;
	}
	
	
	void	
	Dict::matchSign(const char * sign, vector<char*> &w) const
	{
		int pos, npos, slen;
	
		slen = strlen(sign);
	
		/* binary search dictionary for matching signature. */
		pos = bsearch(sign, slen);
	
		if (pos != -1) {
			/* 
			 * search for all remaining duplicate signatures.
			 * the duplicates can be in contiguous locations
			 * on both the left and right sides of the index 
			 * returned by bsearch(). 
			 */
			npos = pos - 1;
			/* scan left */
			while (npos >= off[slen].start && strcmp7_eq(dictarr[npos].sign, sign)) {
				/* copy the word pointers into w[] */
				w.push_back(dictarr[npos--].word);
			}
			/* scan right */
			do {
				w.push_back(dictarr[pos].word);
			} while (++pos < off[slen].end && strcmp7_eq(dictarr[pos].sign, sign));
		}
	}
	
	/* sort by strlen() and then lexicographically */
	static int
	dictSortcmp(const void *p1, const void *p2)
	{
		const char *s1 = ((const dict_ent *)p1)->sign;
		const char *s2 = ((const dict_ent *)p2)->sign;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	static int
	wordSortcmp(const void *p1, const void *p2)
	{
		const char *s1 = (const char *)p1;
		const char *s2 = (const char *)p2;
		int r;
		
		r = strlen(s1) - strlen(s2);
		if (r == 0) 
			return strcmp(s1, s2);
		else 
			return r;
	}
	
	static int
	precompSortcmp(const void *p1, const void *p2)
	{
		const precomp_ent *s1 = (const precomp_ent *)p1;
		const precomp_ent *s2 = (const precomp_ent *)p2;
		
		int r = s1->nsol - s2->nsol;
		if (r == 0) 
			return strcmp(s1->sign, s2->sign);
		else 
			return r;
	}

	void
	Dict::destroy()
	{
		free(dictarr);			dictarr = 0;
		free(words); 			words = 0;
		free(precomparr);		precomparr = 0;
	}
}
